"""
Contains all the arm classes.

Methods:
    * pull(): returns info about the arm
    * info(): pulls the arm
"""

import numpy as np

__all__ = ['BernoulliArm', 'NormalArm', 'LinearArm', 'GeneralArm', 'BanditArm']

class BernoulliArm:
    """
    Arm with reward probabilty from a Bernoulli distribution.

    Positional Arguments:
        * attribute list [mean]
    Attributes:
        * mean: the mean of the random distribution
    Methods:
        * info(): returns info about the arm
        * pull(): pulls the arm

    Examples:

    >>> arm = banditvis.BernoulliArm([0.7])
    >>> arm.info()
    ['Bernoulli', 0.7]
    >>> arm.mean
    0.7
    """
    def __init__(self, attr_list):
        self.mean = attr_list[0]

    def info(self):
        return ["Bernoulli", self.mean]

    def pull(self):
        if np.random.random() > self.mean:
            return 0
        return 1


class NormalArm:
    """
    A NormalArm selects rewards from a normal distribution.

    Positional Arguments:
        * attribute list [mean, variance]
    Attributes:
        * mean: the mean of the random distribution
        * variance: the variance of the random distribution
    Methods:
        * info(): returns info about the arm
        * pull(): pulls the arm

    Examples:

    >>> arm = banditvis.NormalArm([0.7,1])
    >>> arm.info()
    ['Normal', 0.7, 1]
    >>> arm.mean
    0.7
    >>> arm.variance
    1
    """
    def __init__(self, attr_list):
        self.mean = attr_list[0]
        self.variance = attr_list[1]

    def info(self):
        return ["Normal", self.mean, self.variance]

    def pull(self):
        return np.random.normal(self.mean, np.sqrt(self.variance))


class LinearArm:
    """
    A LinearArm selects rewards as a linear function of the arm and the mean.

    Positional Arguments:
        * arm_vector
    Attributes:
        * arm_vec: the arm vector
        * dim: dimension
    Methods:
        * info(): returns info about the arm
        * pull(): pulls the arm
    Caveats:
        * you must manually add an arm_vec method; this is also added when added as an argument to
            a LinBandit object

    Examples:

    >>> arm = banditvis.LinearArm([1,2,3])

    We haven't contributed the mean_vec attribute, so we get an AttributeError:
    >>> arm.info()
    Traceback (most recent call last):
        ...
    AttributeError: 'LinearArm' object has no attribute 'mean_vec'

    Now everything is fine.
    >>> arm.mean_vec = [-1,0,1]
    >>> arm.info()
    ['Linear', [1, 2, 3], [-1, 0, 1]]
    >>> arm.dim
    3
    """
    def __init__(self, arm_vector):
        self.arm_vec = arm_vector
        self.dim = len(arm_vector)

    def info(self):
        return ["Linear", self.arm_vec, self.mean_vec]

    def pull(self):
        return np.inner(self.arm_vec, self.mean_vec) + np.random.normal(0,1)

class GeneralArm:
    def __init__(self, reward_vec):
        self.reward_vec = reward_vec
        self.mean = np.average(reward_vec)

    def info(self):
        return["General", self.mean]

    def pull(self, timestep):
        # fixes indexing
        return self.reward_vec[timestep-1]

class BanditArm:
    def __init__(self, horizon, seed=False):
        self.horizon = horizon
        self.seq = np.array([])
        self.info = []
        self.distr = np.array([])
        if seed:
            np.random.seed(seed)

    def give(self):
        return self.seq, self.distr


    def add(self, **info_list):
        """
        class to build bandit arms
        Positional
        * arm_type

        Arm Type Dependencies:
            * Custom
                * input_
                * prperties

            * Stochastic
                * dist

            * Linear
                * arm_v
                * mean_v

        Sub Type Dependencies:
            * Bernoulli
                * mean
            * Normal
                * mean, variance
            * Beta
                * alpha
                * beta
        """
        try:
            if info_list['armtype'] == 'Custom':
                self._addCustom(**info_list)
            elif info_list['armtype'] == 'Stochastic':
                self._addStochastic(**info_list)
            elif info_list['armtype'] == 'Linear':
                self._addLinear(**info_list)
            else:
                raise AttributeError('BanditArm: {} armtype not recognized'.format(info_list['armtype']))
        except KeyError:
            raise KeyError("You must include an 'armtype' as an argument for BanditArm.add")


    def _add(self, new_vec):
        if self.seq.size:
            self.seq = np.vstack((self.seq, new_vec))
        else:
            self.seq = new_vec

    def _add_distr(self, distr_vec):
        if self.distr.size:
            self.distr = np.vstack((self.distr, distr_vec))
        else:
            self.distr = distr_vec

    def _addStochastic(self, dist=None, mean=0, variance=1, alpha=1, beta=1, **kwargs):
        if dist == 'Bernoulli':
            self._addStochasticBernoulli(mean)
        elif dist == 'Normal':
            self._addStochasticNormal(mean, variance)
        elif dist == 'Beta':
            self._addStochasticBeta(alpha, beta)
        else:
            raise ValueError("BanditArm: improper distribution")
        self._add_distr(np.full((1, self.horizon), mean, dtype=float))
        return None

    def _addCustom(self, input_=None, distr=None, properties={}, **kwargs):
        # input: array_like or file path
        # define custom properties with additional kwargs, overwrites any existion definitions
        # use this to give your stuff nicer names, etc
        new_vec = np.loadtxt(input_)
        if len(new_vec.shape) == 1:
            x, = new_vec.shape
            if x == self.horizon:
                self._add(new_vec)
                self.info.append({**{'armtype': 'Custom', 'source': input_, 'mean': np.mean(new_vec)}, **properties})
            else:
                raise ValueError("BanditArm: array inconsistent with horizon")
        elif len(new_vec.shape) == 2:
            for row in new_vec:
                x, = row.shape
                if x == self.horizon:
                    self._add(row)
                    self.info.append({**{'armtype': 'Custom', 'source': input_, 'mean': np.mean(row)}, **properties})
                else:
                    raise ValueError("BanditArm: array inconsistent with horizon")
        else:
            raise ValueError("BanditArm: unrecognized input_ shape")
        if distr:
            self._add_distr(distr)
        else:
            self._add_distr(new_vec)
        return None

    def _addLinear(self, arm_v=None, mean_v=None, dist=None, noise=None, mean=0, variance=1, **kwargs):
        # support for vector noise coming later
        if dist == 'Normal':
            self._addLinearNormal(arm_v, mean_v, noise, mean, variance)

        elif dist == 'Bernoulli':
            self._addLinearBernoulli(arm_v, mean_v, noise, mean, variance)
        self._add_distr(np.full((1, self.horizon), mean, dtype=float))
        return None


    def _addStochasticNormal(self, mean, variance):
        self._add(np.random.normal(mean, np.sqrt(variance), size=self.horizon))
        self.info.append({'armtype': 'Stochastic Normal', 'mean': mean, 'variance': variance})
        return None

    def _addStochasticBernoulli(self, mean):
        self._add(np.random.binomial(1, mean, size=self.horizon))
        self.info.append({'armtype': 'Stochastic Bernoulli', 'mean': mean})
        return None

    def _addStochasticBeta(self, alpha, beta):
        self._add(np.random.beta(alpha, beta, size=self.horizon))
        self.info.append({'armtype': 'Stochastic Beta', 'alpha': alpha, 'beta': beta})
        return None

    def _addLinearBernoulli(self, arm_v, mean_v, noise, mean, variance):
        if noise == 'scalar':
            # scalar noise just extends each inner product by some noise, preserving direction
            self._add(np.inner(self.arm_vec, self.mean_vec) + np.random.bernoulli(1,mean, size=self.horizon))
        # elif noise == 'vector':
            # vector noise
            # self._add(np.inner(self.arm_vec, self.mean_vec + np.random.normal(mean, variance, size=self.horizon)))
        else:
            raise AttributeError('BanditArm: {} noise not recognized'.format(noise))
        self.info.append({'armtype': 'Linear Bernoulli', 'arm_vector': arm_v, 'mean_vector': mean_v, 'noise': noise, 'mean': mean})
        return None

    def _addLinearNormal(self, arm_v, mean_v, noise, mean, variance):
        if noise == 'scalar':
            # scalar noise just extends each inner product by some noise, preserving direction
            self._add(np.inner(self.arm_vec, self.mean_vec) + np.random.normal(mean, variance, size=self.horizon))
        # elif noise == 'vector':
            # vector noise
            # self._add(np.inner(self.arm_vec, self.mean_vec + np.random.normal(mean, variance, size=self.horizon)))
        else:
            raise AttributeError('BanditArm: {} noise not recognized'.format(noise))
        self.info.append({'armtype': 'Linear Normal', 'arm_vector': arm_v, 'mean_vector': mean_v, 'noise': noise, 'mean': mean, 'variance': variance})
        return None



"""
Entry point for the banditvis script installed along with this package.


Usage:
    banditvis [options] input
    banditvis --default=<source>

    banditvis -h | --help
    banditvis -V | --version

Positional:
  input                 The path to your .txt input file.

Optional:
  -D, --delete          Delete intermediate data files.
  --data=<directory>    The path to the data directory, defaults to the
                          current directory.
  --out=<directory>     The path to the output location, defaults to the
                          current directory.
  -v, --verbose         Display additional information.

Other:
  --default=<source>    Source defaults from a specified file.
"""
from . import __version__
from .manager import run
import time
import sys
import os
import yaml
import argparse
import pickle
from pprint import pprint

def get_usage():
    return '\n\n\n'.join(__doc__.split('\n\n\n')[1:])

def get_args():
    """
    Command line argument parser. Parses the arguments
    """
    # nargs='?' means 0 or 1 argument
    parser = argparse.ArgumentParser(prog='banditvis')
    parser.add_argument("input", nargs='?',
        help="The path to your .txt input file.")
    parser.add_argument("--data", nargs='?',
        help="The path to the data directory, defaults to the current directory.")
    parser.add_argument("--out", nargs='?',
        help="The path to the output location, defaults to the current directory.")
    parser.add_argument("-D", "--delete", action='store_true',
        help="Delete intermediate data files.")
    parser.add_argument("--default", nargs='?',
        help="Source defaults from a specified file.")
    parser.add_argument("-v","--verbose", action='store_true',
        help="Display additional information.")

    # check for certain arguments
    if (("-h" or "--help") in sys.argv[1:]) or (len(sys.argv) == 1):
        print(get_usage())
        sys.exit(0)
    elif ("-V" or "--version") in sys.argv[1:]:
        print("banditvis ({}) -- installed at {}".format(__version__, os.path.dirname(__file__)))
        sys.exit(0)

    # parse the arguments
    else:
        unmod_args = vars(parser.parse_args())
        args = {key:unmod_args[key] for key in unmod_args.keys() if unmod_args[key] is not None}
        if 'default' in args.keys():
            file_name = args['default']
            try:
                with open(file_name, 'r') as file:
                    user_defaults = yaml.load(file)
                with open(os.path.dirname(__file__) + '/user_defaults.pkl', 'wb') as handle:
                    pickle.dump(user_defaults, handle)
                if args['verbose']:
                    print("Successfully loaded default file! Results:\n")
                    pprint(user_defaults)
                    print("\n")
                else:
                    print("Successfully loaded default file!")

                sys.exit(0)
            except FileNotFoundError:
                print("Default file '{}' not found.".format(file_name))
                sys.exit(1)
        # remove any None values
        return args

def main():
    run(**get_args())

if __name__ == "__main__":
    main()




"""
BanditVis

Simulate and visualize Bandit algorithms.

:copyright: (c) 2016 by Alex Rutar.
:license: MIT, see LICENSE for more details.
"""

__version__ = '0.2'

from banditvis.core import StoBandit, LinBandit, Algorithm, Bandit
from banditvis.algorithms import *
from banditvis.arms import BernoulliArm, NormalArm, LinearArm, BanditArm
from banditvis.simulation import Simulation


import pickle
import os


CORE_DEFAULTS = {
    'out': "hi",
    'Animate': False,
    'FPS': 20,
    'HelpLines': True,
    'InputData': False,
    'LevelCurves': True,
    'Multiprocess': 1,
    'NoAxesTick': False,
    'Normalized': False,
    'PlotSave': "temp.pdf",
    'PlotTitle': False,
    'DeleteData': False
}

def load_user():
    with open(os.path.dirname(__file__) + '/user_defaults.pkl', 'rb') as handle:
        _user = pickle.load(handle)
    return _user


from sys import stdout
import time
from .algorithms import *
from .arms import *
from .core import *

import numpy as np
from pprint import pprint


class Simulation:

    """
    Class to combine the Bandit - Algorithm interaction.

    Args:
        bandit_inst: first argument is an instance of a Bandit (StoBandit, LinBandit, AdvBandit)
        algorithm_inst: second argument is an instance of an Algorithm

    Methods:
        ()
        reset()

    TODO:
    * what else to track? total reward, calculate regret, ...

    """

    def __init__(self, bandit_inst, algorithm_inst):
        """
        Initializes the Simulation class, and passes certain attributes so that
        Bandit objects and Algorithm objects have acces to certain information.
        """
        self.alg = algorithm_inst
        self.bandit = bandit_inst
        self.alg.bandit = bandit_inst
        self.total_regret = 0
        self.iterations = 0
        self.arm_info = [each_arm.info() for each_arm in self.bandit.arms]


    def reset(self):
        """
        Resets the Simulation class the Bandit subclass.
        """
        self.iterations = 0
        self.total_regret = 0
        self.bandit.reset()


    def _print_info(self):
        """
        Prints out information about the Simulation.

        The 'try' statement attempts to print out the .__name__ attribute of
        functions, rather than the object itself; if it fails, it falls back to
        just printing the object itself
        """
        print("\nAverage Regret: {0}".format(
            self.total_regret / self.iterations))

        for i in self.alg.var_dict:
            try:
                print("{0} : {1}".format(i, self.alg.var_dict[i].__name__))
            except AttributeError:
                print("{0} : {1}".format(i, self.alg.var_dict[i]))

        print("{0} Arms: {1}".format(self.bandit.n_arms, self.arm_info))
        print("Horizon: {0}".format(self.horizon))
        print ("Total runtime: {0} seconds".format(self.runtime))
        print ("Average runtime per iteration: {0} seconds".format(
            self.runtime / self.iterations))
        print("\n" + "-"*50 + "\n")
        return None


    def runStep(self, num, horizon):
        """
        Runs the Simulation for one step.

        This method is called during animations.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for m in range(num):
            self.bandit.pullArm(self.alg.giveArm())
        return None


    def runView(self, horizon):
        """
        Runs a single iteration to a horizon of n and prints out information.

        runView is especially useful for debugging bandit algorithms or seeing
        everything in significant detail as the bandit runs.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        start_time = time.clock()
        for m in range(self.bandit.n_arms):
            self.bandit.pullArm(m)
            self.bandit.fullInfo()
        for m in range(horizon - self.bandit.n_arms):
            self.bandit.pullArm(self.alg.giveArm())
            self.bandit.fullInfo()
        stop_time = time.clock()

        self.total_regret += self.bandit.giveRegret()
        self.iterations += 1
        self.bandit.reset()
        self.runtime = stop_time - start_time
        self._print_info()
        return None


    def runStandard(self, horizon, cycles):
        """
        A standard experiment-running method.

        Runs the bandit for a certain number of cycles each to a certain
        horizon, and prints out the result on screen.

        It uses _print_info to print information about the bandit itself upon
        termination.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        print("-"*50 + "\n")
        start_time = time.clock()

        for i in range(cycles):
            for i in range(self.bandit.n_arms):
                self.bandit.pullArm(i)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1
            self.bandit.reset()

            stdout.write(
                "\r----------  "
                "{0:d} out of {1} cycles"
                "----------".format(self.iterations, cycles))
            stdout.flush()

        stop_time = time.clock()
        self.runtime = stop_time - start_time
        self._print_info()
        return None


    def runHist(self, horizon, cycles, out_file):
        """
        Produces Histogram - like data.

        Every cycle, prints the bandit regret it out in the file, followed by a
        newline. These data files are used by many other methods containing
        hist in their name.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for i in range(cycles):
            for j in range(self.bandit.n_arms):
                self.bandit.pullArm(j)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1

            with open(out_file, "a") as outfile:
                outfile.write("{0}\n".format(self.bandit.giveRegret()))
            self.bandit.reset()
        return None


    def runVar(self, horizon, cycles, out_file):
        """
        Produces Variable - like data.

        Runs the simulation for a certain number of cycles, then appends that
        data to the text file
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for i in range(cycles):
            for i in range(self.bandit.n_arms):
                self.bandit.pullArm(i)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1

            self.bandit.reset()

        with open(out_file, "a") as outfile:
            outfile.write("{0}\n".format(self.total_regret/self.iterations))
        return None

    def selfCheck(self):
        """
        Checks the simulation for errors
        """
        errors = ""
        return errors

class ObjectDict:
    ArmDict = {
        'Bernoulli' : BernoulliArm,
        'Normal' : NormalArm,
        'Linear' : LinearArm
    }
    AlgDict = {
        'var_dict' : {},
        'random' : random,
        'greedy' : greedy,
        'greedy_ep' : greedy_ep,
        'UCB' : UCB,
        'UCB_KL' : UCB_KL,
        'TS_Beta' : TS_Beta,
        'TS_Gauss' : TS_Gauss,
        'Bayes_Gauss' : Bayes_Gauss,
        'UCB_Lin' : UCB_Lin,
        'TS_Lin' : TS_Lin
    }
    IndexDict = {
        'B1' : B1,
        'B2' : B2,
        'B3' : B3,
        'B4' : B4,
        'B5' : B5,
        'B6' : B6,
        'B7' : B7,
        'B8' : B8,
        'B9' : B9
    }


def ReMapSim(sim_dict):
    """
    Takes a simulation dictionary and turns string names into objects. It uses
    the ObjectDict class in order to create references to the actual objects.
    """
    for alg_key in list(sim_dict['Algorithm'].keys()):
        if alg_key == 'incr':
            sim_dict['Algorithm'][alg_key] = \
                ObjectDict.IndexDict[sim_dict['Algorithm'][alg_key]]
        elif alg_key == 'algtype':
            sim_dict['Algorithm'][alg_key] = \
                ObjectDict.AlgDict[sim_dict['Algorithm'][alg_key]]

    sim_dict['Algorithm'] = Algorithm(**sim_dict['Algorithm'])
    sim_dict['arm_object_list'] = [ObjectDict.ArmDict[arm[0]](arm[1])
        for arm in sim_dict['Bandit']['ArmList']]

    # different parsing for Linear and non-Linear bandits
    if sim_dict['Bandit']['ArmList'][0][0] in ('Linear'):
        sim_dict['vector_mean'] = sim_dict['Bandit']['MeanVector']
        sim_dict['Bandit'] = LinBandit(
            sim_dict['arm_object_list'],
            sim_dict['vector_mean'],
            normalized=sim_dict['Normalized'])
        del[sim_dict['vector_mean']]
    elif sim_dict['Bandit']['ArmList'][0][0] in ('Bernoulli', 'Normal'):
        sim_dict['Bandit'] = StoBandit(sim_dict['arm_object_list'])

    sim_dict['Simulation'] = Simulation(
        sim_dict['Bandit'],
        sim_dict['Algorithm'])

    # clean up residual keys
    del[sim_dict['Bandit']]
    del[sim_dict['Algorithm']]
    del[sim_dict['arm_object_list']]

    return None



import sys
import multiprocessing as mp
import re

import numpy as np
import yaml
from datetime import datetime
from pprint import pprint

from .helper import msplit
from .formatting import bcolors
from .defaults import load_user, CORE_DEFAULTS

def Parse(user_file, **arg_dict):
    """
    Loads the user_file using YAML and then checks the dictionary for
    consistency. Once Parse is completed, it either exits with errors or
    returns a checked core_dict.
    """

    # Try to load the file and catch YAML syntax errors for nicer formatting
    try:
        # merge the two dicts
        core_dict = CoreDict(yaml.load(open(user_file, 'r')), **arg_dict)
    except yaml.parser.ParserError as e:
        # pull some information out of the ParserError
        location = (str(e).split("\n")[1].split(",")[1] + str(e).split("\n")[1].split(",")[2])[1:]
        e_msg = str(e).split("\n")[2]
        sys.exit(bcolors.FAIL + "\nYAML Syntax Error " + bcolors.ENDC
            + "in {}:\n  - {}\n".format(location, e_msg)
        )
    except yaml.scanner.ScannerError as e:
        # pull some information out of the ScannerError
        location = (str(e).split("\n")[1].split(",")[1]
            + str(e).split("\n")[1].split(",")[2]
            )[1:]
        e_msg = str(e).split("\n")[0]
        sys.exit(bcolors.FAIL + "\nYAML Syntax Error " + bcolors.ENDC
            + "in {}:\n  - {}\n".format(location, e_msg)
            )

    # creates a simulation dictionary list for easier access
    core_dict['sim'] = []
    for sim_key in list(core_dict.keys()):
        if sim_key[:10] == 'Simulation':
            core_dict['sim'].append(CoreDict(core_dict[sim_key]))
            del(core_dict[sim_key])

    # perform the dictionary check, and get the errors
    core_dict, errors = DictCheck(core_dict)

    # if there are errors, print them
    if errors:
        sys.exit(bcolors.FAIL
            + "\n" + " ERROR LIST ".center(100, "-") + "\n" + bcolors.ENDC
            + self.errors + "\n\n")
    else:
        print(bcolors.OKGREEN + "\n" + "No Errors!".center(100, "-") + "\n" + bcolors.ENDC)

    warnings = core_dict.warnings()
    if warnings:
        print(bcolors.WARN
            + "\n" + " WARNING LIST ".center(100, "-") + "\n" + bcolors.ENDC
            + "".join(["- {}\n".format(item) for item in warnings]))
        user_response = input("Press Enter to continue, and <any key> + Enter to abort. ")
        if user_response:
            sys.exit(0)

    else:
        print(bcolors.OKGREEN + "\n" + " No Warnings! ".center(100, "-") + "\n" + bcolors.ENDC)

    # done!
    return core_dict


def DictCheck(core_dict):
    """
    The checking function, uses _check class to perform these checks nicely.

    Defaults are built first, followed by checks based on each init file and
    other specifics. The full list of checking functions with explanations
    can be found in the _check class.
    """
    check = Check(core_dict)
    if core_dict['InputData']:
        check.SimExist('label')
        check.Save()
        check.Title()


    elif core_dict['init'] == 'Variable':
        check.Save()

        check.Conflict('horizon')
        check.Conflict('cycles')
        check.SimExist('label')

        check.Args()
        check.Linecount()
        check.Folder()
        check.Title()


    elif core_dict['init'] == 'Histogram':
        check.Save()

        check.Conflict('horizon')
        check.Conflict('cycles')
        check.SimExist('label')

        check.Bins()
        check.Linecount()
        check.Folder()
        check.Title()


    elif core_dict['init'] == 'Visualize':
        pass


    else:
        sys.exit(bcolors.FAIL
            + '\n\nIf you get this message, something wend badly wrong...\n'
            + bcolors.ENDC
        )
    return check.give()


class CoreDict(dict):
    """
    Subclasses dict to change behaviour with missing key (allows setting of deafults easily) and
    some other custom methods.
    """
    def __init__(self, in_dict, core_defaults=CORE_DEFAULTS, **arg_dict):
    # def __init__(self, in_dict, core_defaults=CORE_DEFAULTS, user_defaults=load_user(), **arg_dict):
        dict.__init__(self, in_dict)

        # precedence: arg_dict, then user_defaults, then core_defaults
        self.default = {**core_defaults, **arg_dict}
        self.ignore = {'InputData', 'DataFolder', 'Animate'}
        self.warning_list = []

    def __missing__(self, key):
        try:
            if key not in self.ignore:
                self.warning_list += ["{} missing, defaulting to '{}'".format(
                    key,
                    self.default[key])
                ]
            return self.default[key]
        except KeyError:
            raise KeyError("CoreDict '{}' key missing in both the dict and the default dict.".format(key))

    def warnings(self):
        return [item for item in list(set(self.warning_list))]

    def state(self):
        print("\n" + " Completed Core Dict ".center(100, "-") + "\n")
        pprint(self)
        print("\n" + " Default Dict ".center(100, "-") + "\n")
        pprint(self.default)
        print("\n" + " End State ".center(100, "-") + "\n")
        return None


class Check:
    """
    A class which checks a dictionary for errors. It contains a large number of
    methods which are convenient / reusable when parsing dictionary objects, and also custom methods
    to do variable formatting.

    Positional Arguments:
        * core_dict: a core_dict or dict-like object
    Attributes:
        * core_dict: the core_dict
        * errors: a list of errors
    Methods:
        * give(): returns the core_dict and a formatted of error string
        * Visual(name): (Currently doesn't do anything)
        * Conflict(name): checks for conflict / multiple declarations of 'name'
        * Exist(name): checks if name exists in the dict
        * SimExist(name): checks if name exists in every simulation sub dictionary
        * Save(): checks for the save title, as well as the save output folder
        * Bins(): figure out appropriate bin allocation for histogram plots
        * Args(): check the args for consistency / proper declaration
        * Folder(): creates a data_folder name and prepends the path to it
        * Linecount(): determines how many lines will be printed to data files
        * Title(): creates a title if none exists and formats an existing one
    """

    def __init__(self, core_dict):
        self.core_dict = core_dict
        self.errors = []


    def give(self):
        return (self.core_dict, "\n".join(set(self.errors)))  # removes repeats from self.errors


    def Visual(self, name):
        if core_dict['visual'] == 'confidence':
            pass
        elif core_dict['visual'] == 'ellipse':
            pass
        elif core_dict['visual'] == 'distribution':
            pass
        else:
            pass


    def Conflict(self, name):
        """
        Checks if 'name' exists in the sim dictionaries and in the overall
        dictionary; looks for conflict / multiple declarations of name. If
        there are no errors, it moves every instance of the declaration to
        the simulation subdictionaries for easier access later.
        """
        checklist = \
            {'low_{}'.format(name) : [False] * len(self.core_dict['sim']),
            'top_{}'.format(name) : [False]}

        if name in self.core_dict.keys():
            checklist['top_{}'.format(name)] = True

        for i, sim_dict in enumerate(self.core_dict['sim']):
            if name in sim_dict.keys():
                checklist['low_{}'.format(name)][i] = True

        if not checklist['top_{}'.format(name)]:
            cntr = 0
            for item in checklist['low_{}'.format(name)]:
                if not item:
                    cntr += 1
            if cntr != 0:
                self.errors += ["- ({0}) was not declared at the top level, and you are missing a"
                    " ({0}) declaration in ({1}) simulation(s)".format(name, cntr)]
        else:
            cntr = 0
            for item in checklist['low_{}'.format(name)]:
                if item:
                    cntr += 1
            if cntr != 0:
                self.errors += ["- ({0}) was declared at the top level, but "
                "you have ({0}) declarations in ({1}) simulation(s).".format(
                    name, cntr)]
            else:
                for sim_dict in self.core_dict['sim']:
                    sim_dict['{}'.format(name)] = \
                        self.core_dict['{}'.format(name)]
                del self.core_dict['{}'.format(name)]
        return None


    def Exist(self, name):
        """
        Checks if 'name' exists in the core_dict
        """
        if name not in self.core_dict.keys():
            self.errors += ["- You are missing a top-level ({}) declaration.".format(
                name)]
        else:
            pass
        return None


    def SimExist(self, name):
        """
        Checks if 'name' exists in every single simulation sim_dictionary, and
        adds an error if it doesn't
        """
        checklist = {'low_{}'.format(name) : [False] * len(self.core_dict['sim'])}
        for i, sim_dict in enumerate(self.core_dict['sim']):
            if name in sim_dict.keys():
                checklist['low_{}'.format(name)][i] = True
        if (checklist['low_{}'.format(name)]
            != [True] * len(self.core_dict['sim'])):
            self.errors += ["- You are missing at least one simulation ({}) declaration.".format(
                name)]
        return None


    def Save(self):
        """
        Checks the save format for .pdf or .png. It returns an eror otherwise.
        """
        if (not self.core_dict['PlotSave'].endswith(".pdf")
            and not self.core_dict['PlotSave'].endswith(".png")):
            self.errors += ["- PlotSave: You can only save a plot as a '.pdf' or '.png'. PDF is "
                "the recommended file type for image quality."]
        if self.core_dict.default['out']:
            self.core_dict['PlotSave'] = self.core_dict.default['out'] + "/" + self.core_dict['PlotSave']
            self.core_dict.default['PlotSave'] = self.core_dict.default['out'] + "/" + self.core_dict.default['PlotSave']
        return None


    def Bins(self):
        """
        Determines how many bins the histogram should have based on the proper-
        ties of the bandit.
        """
        self.core_dict['bins'] = []
        for sim_dict in self.core_dict['sim']:
            if sim_dict['Bandit']['ArmList'][0][0] == 'Linear':
                mean_list = [np.inner(arm[1], sim_dict['Bandit']['MeanVector'])
                    for arm in sim_dict['Bandit']['ArmList']]
            else:
                mean_list = [arm[1] for arm in sim_dict['Bandit']['ArmList']]
            self.core_dict['bins'].append(np.amax(
                mean_list - np.amin(mean_list)) * sim_dict['horizon'])
        return None


    def Args(self):
        """
        When using the Variable class, args can be declared in multiple ways.
        This builds a single list of args out of the inputs to be used later.
        """
        checklist = {'linspace': False, 'args': False}
        if ('domain' in self.core_dict['Var'].keys()
            and 'samples' in self.core_dict['Var'].keys()):
            checklist['linspace'] = True

        if 'arg_list' in self.core_dict['Var'].keys():
            checklist['args'] = True

        if checklist['linspace'] and checklist['args']:
            self.errors += ["- Declare either (domain and samples) or (args), not both."]

        elif not checklist['linspace'] and not checklist['args']:
            self.errors += ["- You must declare either (domain and samples) or (args)."]
        else:
            try:
                self.core_dict['arg_list'] = np.linspace(
                    self.core_dict['Var']['domain'][0],
                    self.core_dict['Var']['domain'][1],
                    self.core_dict['Var']['samples'])
            except KeyError:
                pass
        return None


    def Folder(self):
        """
        Additional changes to be done after checking
        """
        self.core_dict.default.setdefault('DataFolder', "{}".format(
            datetime.strftime(datetime.now(), '%Y-%m-%d %H_%M_%S')))
        if self.core_dict.default['data']:
            self.core_dict.default['DataFolder'] = self.core_dict.default['data'] + "/" + self.core_dict.default['DataFolder']
        return None


    def Linecount(self):
        if self.core_dict['init'] == 'Histogram':
            self.core_dict['total_lines'] = sum(sim_dict['cycles'] for sim_dict in self.core_dict['sim'])
        elif self.core_dict['init'] == 'Variable':
            self.core_dict['total_lines'] = len(self.core_dict['arg_list']) * len(self.core_dict['sim'])
        return None


    def Title(self):
        if not self.core_dict['PlotTitle']:
            self.core_dict['PlotTitle'] = msplit(self.core_dict['PlotSave'], "/", ".")[-1] + "\n"
        else:
            self.core_dict['PlotTitle'] += "\n"

        return None



import unittest

import banditvis

class FirstTest(unittest.TestCase):
    def test_is_string(self):
        s = "superhappystring"
        self.assertTrue(isinstance(s, int))

if __name__ == "__main__":
    unittest.main()


import matplotlib as mpl
import matplotlib.pyplot as plt

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARN = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''

class cmap_colors:
    sequential1 = [plt.cm.get_cmap(name) for name in ['Blues', 'Greens', 'Reds', 'BuGn',
        'BuPu','GnBu', 'Greys', 'Oranges',
        'OrRd','PuBu', 'PuBuGn', 'PuRd',
        'Purples', 'RdPu', 'YlGn', 'YlGnBu']]
    sequential2 = [plt.cm.get_cmap(name) for name in ['afmhot', 'autumn', 'bone', 'cool',
        'copper', 'gist_heat', 'gray', 'hot',
        'pink', 'spring', 'summer', 'winter']]

class mpl_defaults:
    def ani():
        """
        Default style used for all animations
        """
        plt.style.use('bmh')
        mpl.rcParams['toolbar'] = 'None'
        plt.rcParams['font.size'] = 12
        plt.rcParams['xtick.labelsize'] = 10
        plt.rcParams['ytick.labelsize'] = 10
        plt.rcParams['legend.fontsize'] = 12
        plt.rcParams['figure.titlesize'] = 12
        plt.rcParams['font.size'] = 10
        plt.rcParams['axes.labelsize'] = 12
        plt.rcParams['axes.labelweight'] = 'bold'
        plt.rcParams['legend.edgecolor'] = '#e6e6e6'
        plt.rcParams['legend.facecolor'] = '#ffffff'
        plt.rcParams['legend.numpoints'] = 1

        fig = plt.figure(figsize=(14,9), facecolor='white')
        ax = fig.add_subplot(111, facecolor='white')

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)

        return (fig, ax)
    def plot():
        """
        Default style used for all plots
        """
        plt.style.use('bmh')
        plt.rcParams['font.size'] = 10
        plt.rcParams['xtick.labelsize'] = 8
        plt.rcParams['ytick.labelsize'] = 8
        plt.rcParams['legend.fontsize'] = 10
        plt.rcParams['figure.titlesize'] = 12
        plt.rcParams['font.size'] = 10
        plt.rcParams['axes.labelsize'] = 10
        plt.rcParams['axes.labelweight'] = 'normal'
        plt.rcParams['lines.linewidth'] = 1
        plt.rcParams['axes.titlesize'] = 'medium'
        plt.rcParams['axes.titleweight'] = 'medium'
        plt.rcParams['legend.edgecolor'] = '#e6e6e6'
        plt.rcParams['legend.facecolor'] = '#ffffff'

        fig = plt.figure()
        ax = fig.add_subplot(111)

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)

        return (fig, ax)



import copy
import os
import numpy as np
from .simulation import ReMapSim
from pprint import pprint
import time

def _fix_vars(obj, variable):
    """
    Recursively searches the nested sim dict structure, and attempts to find
    every instance of '&&', replace it with some variable, and evaulate that
    string. It does this by recursively building a copy of the object if it
    contains dict or list substructures. Note that this function will not
    search in any structures other than lists or dicts.
    """
    if isinstance(obj, dict):
        return {key: _fix_vars(val, variable) for key, val in obj.items()}
    elif isinstance(obj, list):
        return [_fix_vars(key, variable) for key in obj]
    elif isinstance(obj, str):
        if '&&' in obj:
            return eval(obj.replace("&&", str(variable)))
        else:
            return obj
    return obj

def HistData(i, sim_dict, data_folder="."):
    """
    Generates Histogram data.

    Runs the bandit algorithms to create files containing regret data for each
    element of core_dict['sim']
        * builds a temporary dictionary which is a deep copy of the correspon-
            ding simulation dictionary
        * uses ReMapSim to build a Simulation object
        * runs that Simulation object using runHist
        * saves that data in the specified folder
    """
    file_name = "{}/data{}.txt".format(data_folder, i)
    temp_dict = copy.deepcopy(sim_dict)
    ReMapSim(temp_dict)
    sim = temp_dict['Simulation']

    sim.runHist(
        temp_dict['horizon'],
        temp_dict['cycles'],
        file_name)

    return None


def VarData(i, sim_dict, data_folder=".", iter_list = []):
    """
    Generates Variable data.

    VarData runs the bandit algorithms to create files containing regret data
    for each element of core_dict['sim']. It:
        * builds a temporary dictionary which is a deep copy of the correspond-
            ing simulation dictionary
        * uses ReMapSim to build a Simulation object
        * runs that Simulation object using runVar
        * saves the output data to sequentially generated save files

    The inner loop works as follows:
        * temp_dict is made as a copy
        * every instance of '#' is replaced with num
        * the bandit algorithm is run using the replaced string
    """
    for num in iter_list:  # now num is the variable
        temp_dict = _fix_vars(sim_dict, num)

        ReMapSim(temp_dict)
        sim = temp_dict['Simulation']

        file_name = "{}/data{}.txt".format(data_folder, i)
        sim.runVar(
            temp_dict['horizon'],
            temp_dict['cycles'],
            file_name
        )

    return None


import argparse
import os

# parser = argparse.ArgumentParser(description="A command line interface for the banditvis module.")

def safe_save(file_name):
    """Save a plot safely."""
    if file_name.split("/")[-1][:4] == 'temp':
        return file_name
    elif not os.path.isfile(file_name):
        return file_name
    else:
        for i in range(100000):
            if os.path.isfile(file_name.split(".")[0]
                + "{}.".format(i)
                + file_name.split(".")[1]):
                pass
            else:
                return(file_name.split(".")[0]
                    + "{}.".format(i)
                    + file_name.split(".")[1])

def msplit(text, *limiters):
    """
    Like split, but separates a string at multiple places.

    Args:
        text: a string that you want to split
        *limiters: strings that you want to split at
    Returns:
        A list of strings

    Examples:

    >>> msplit("example test string here!", 'p', 'g')
    ['exam', 'le test strin', ' here!']

    """
    for limit in limiters:
        text = text.replace(limit, '$!!7=4+[4}|[2')
    return text.split('$!!7=4+[4}|[2')


from setuptools import setup, find_packages
def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='banditvis',
    version='0.6',
    description='A library for simulating and visualizing bandit algorithms.',
    long_description=readme(),
    url='https://github.com/alexrutar/banditvis',
    author='Alex Rutar',
    author_email='arutar@uwaterloo.ca',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License'
    ],
    keywords='bandit machine learning',
    license='MIT',
    packages=find_packages(exclude=['docs', 'tests', 'checklist.py']),
    install_requres=[
        'numpy',
        'pyyaml',
        'scipy',
        'matplotlib'
    ],
    package_data={'': ['LICENSE', 'README.md', 'banditvis/user_defaults.pkl']},
    include_package_data=True,
    entry_points={'console_scripts': ['banditvis = banditvis.__main__:main']},
    zip_safe=False
)



"""
The manager module for the command line tools. It does multiprocess management, as well as
"""
import multiprocessing as mp
import threading as th
from pprint import pprint
import sys
import time
import signal
import os

from .parse import Parse
from .plot import *
from .data import *
from .animation import *
from .formatting import bcolors

def _checkInput(**arg_dict):
    """
    Gets the input from the user commmand and checks it for potential errors.
    """
    if arg_dict['input'].endswith('.txt'):
        pass
    else:
        sys.exit("ERROR: You need to input a .txt file.")
    try:
        input_file = arg_dict['input']
        del(arg_dict['input'])
        core_dict = Parse("{}".format(input_file), **arg_dict)
    except FileNotFoundError:
        sys.exit("ERROR: The file '{}' you tried to input doesn't exist in the current directory.".format(input_file))
    return core_dict

def _init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class _statusThread(th.Thread):
    def __init__(self, core_dict):
        super(_statusThread, self).__init__()
        self.core_dict = core_dict
        self.daemon = True
    def run(self):
        n_files = len(self.core_dict['sim'])
        current = [0] * n_files
        finished = 0
        total = self.core_dict['total_lines']
        while finished != total:  # TODO fix this to have a termination
            time.sleep(0.5)
            for i in range(n_files):
                try:
                    with open("{}/data{}.txt".format(self.core_dict['DataFolder'], i), 'r') as file:
                        current[i] = sum(1 for line in file) - 1
                except FileNotFoundError:
                    pass
            finished = sum(current)
            sys.stdout.write("\r" + " {:2.0f} % complete ".format(finished*100/total).center(100, "-"))
            sys.stdout.flush()


def run(**arg_dict):
    core_dict = _checkInput(**arg_dict)
    pool = mp.Pool(core_dict['Multiprocess'], _init_worker)
    # ----------------------------------------------------------------------------------------------
    if arg_dict['verbose']:
        print(bcolors.OKBLUE)
        print("\n" + " Argument Dict ".center(100, "-") + "\n")
        pprint(arg_dict)
        core_dict.state()  # displays the core_dict
        print(bcolors.ENDC)

    try:
        if core_dict['init'] == 'Histogram':
            os.mkdir(core_dict['DataFolder'])
            status = _statusThread(core_dict)
            status.start()
            if not core_dict['InputData']:
                pool.starmap(HistData, [(i, sim_dict, core_dict['DataFolder'])
                    for i, sim_dict in enumerate(core_dict['sim'])])
                pool.close()
                pool.join()
                if core_dict['Animate']:  # TODO in general fix this thing
                    while True:
                        try:
                            HistAnimation(core_dict)
                            break
                        except FileNotFoundError:
                            time.sleep(0.3)
            else:
                pass
            HistPlot(core_dict)

        elif core_dict['init'] == 'Variable':
            os.mkdir(core_dict['DataFolder'])
            status = _statusThread(core_dict)
            status.start()
            if not core_dict['InputData']:
                pool.starmap(VarData, [(i, sim_dict, core_dict['DataFolder'], core_dict['arg_list'])
                    for i, sim_dict in enumerate(core_dict['sim'])])
                pool.close()
                pool.join()
            else:
                pass
            VarPlot(core_dict)

        elif core_dict['init'] == 'Visualize':
            if core_dict['visual'] == 'ellipse':
                EllipseAnimation(core_dict)
            elif core_dict['visual'] == 'confidence':
                ConfAnimation(core_dict)
            elif core_dict['visual'] == 'distribution':
                DistAnimation(core_dict)

        # ------------------------------------------------------------------------------------------
        # POST PROCESSING
        # ------------------------------------------------------------------------------------------

        print("\n\n" + bcolors.OKGREEN + " Done! Saved to {} ".format(core_dict['PlotSave']).center(100,"-") + bcolors.ENDC + "\n")

        if core_dict['delete'] or core_dict['DeleteData']:
            path = core_dict['DataFolder']
            for i in range(len(core_dict['sim'])):
                os.remove(path + "/data{}.txt".format(i))
            os.rmdir(path)


    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
        print("\r" + " "*150)
        print("\n\n" + bcolors.FAIL + " Keyboard Interrupt! ".center(100,"-") + bcolors.ENDC + "\n")
        sys.exit(1)





"""
This module contains the various Bandit classes {StoBandit, LinBandit, AdvBandit} and the Algorithm
class.
"""

import numpy as np
from numpy.linalg import inv
from pprint import pprint

__all__ = ['StoBandit', 'LinBandit', 'AdvBandit', 'Algorithm']


class StoBandit:

    """
    This bandit class simulates a single-dimensional `Sto`chastic bandit.

    Positional Arguments:
        * arm object
    Attributes:
        * Bandit.arms : list of arms objects within the bandit, which contains:
            * Bandit.arms[n].mean : the mean of the n^th arm
            * Bandit.arms[n].info : info about the n_th arm
        * Bandit.n_arms : number of arms that the Bandit has
        * Bandit.T : a vector storing the number of pulls of each arm
        * Bandit.U : a vector storing the cumulative average of each arm
        * Bandit.U_conf : the adjusted version of U [n] taking into account
            some confidence interval (not always used)
        * Bandit.timestep[0] : the number of timesteps that have passed so far
            (incremented with every arm pull)
            * Note : Bandit.timestep is a vector for use in algorithms
        * Bandit.total_reward : the total reward that the Bandit has
            accumulated so far
        * Bandit.arm_reward : the reward that each arm has so far
        * Bandit.reward : the reward that the Bandit recieved after the
            previous timestep
    Methods:
        * pullArm(arm): pull a specific arm
        * giveRegret(): calculate the current regret
        * reset(): reset all timestep-based properties
        * fullInfo(): print information about the bandit
    """

    def __init__(self, arm_object_list):
        self.arms = arm_object_list
        self.n_arms = len(arm_object_list)
        self.T = np.zeros(self.n_arms, dtype=np.int)
        self.U = np.zeros(self.n_arms)
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        self.arm_reward = np.zeros(self.n_arms, dtype=np.int)
        self.total_reward = 0

        self.mean_list = np.array([self.arms[arm].mean
            for arm in range(self.n_arms)])


    def giveRegret(self):
        return sum(self.T[i] * (np.amax(self.mean_list) - self.arms[i].mean) for i in range(self.n_arms))


    def pullArm(self, arm):
        self.T[arm] += 1
        self.timestep += 1
        self.reward = self.arms[arm].pull()
        self.arm_reward[arm] += self.reward
        self.U[arm] = self.U[arm] + 1/self.T[arm] * (self.reward - self.U[arm])

        return None


    def reset(self):  # resets everything in the bandit for re-use!
        self.T = np.zeros(self.n_arms, dtype=np.int)
        self.U = np.zeros(self.n_arms)
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        self.regret = 0

        return None


    def fullInfo(self):
        print("\n" + "+" + "-"*85 + "+")

        for arm in range(self.n_arms):
            print("| Arm {0}: mean ({1}) average reward ({2:f}), confidence "
                "({3:f}), called ({4}) times"
                .format(
                    arm,
                    self.arms[arm].mean,
                    self.U[arm],
                    self.U_conf[arm],
                    self.T[arm]))
        print("|\n| {0} arms, regret ({1}), timestep ({2} / {3})"
            .format(
                self.n_arms,
                self.giveRegret(),
                self.timestep[0],
                self.horizon[0]))

        print("+" + "-"*85 + "+" + "\n")
        return None



class LinBandit:
    """
    The Linear Bandit class is very similar to the original Bandit class, in
    that it contains similar methods which do the same thing as the original
    Bandit. However, it differs in the way in which is stores averages, and
    takes different arm arguments.

    Positional Arguments:
        * list of arm objects
        * vector mean
    Keyword Arguments:
        * normalized: boolean
    Attributes:
        * normalized: boolean for normalized (all vectors length 1)
        * mean: the vector mean
        * n_arms: the number of arms
        * T: the number of pulls of each arm
        * arms: the arm_object_list input
        * dim: dimension
        * arm_vecs: a list of the arm vectors
        * G: the gram matrix, initialized with the identity to ensure
            invertibility
        * U: the system average vector
        * A: the sum of arm vectors chosen multiplied by the scalar reward
            it recieved
        * U_conf: the confidence value for each arm
        * timestep: the timestep
    Methods:
        * pullArm(arm): pull a specific arm
        * giveRegret(): calculate the current regret
        * reset(): reset all timestep-based properties
        * fullInfo(): print information about the bandit
    """
    def __init__(self, arm_object_list, vector_mean, normalized=False):
        self.mean = vector_mean
        self.arms = arm_object_list
        self.arm_vecs = np.array([arm.arm_vec for arm in self.arms])
        self.normalized = normalized

        if self.normalized:
            self.mean = self.mean/np.linalg.norm(self.mean)
            self.arm_vecs = np.array([arm/np.linalg.norm(arm) for arm in self.arm_vecs])

        self.n_arms =  len(arm_object_list)
        self.T = np.zeros(len(arm_object_list), dtype = int)
        self.dim = arm_object_list[0].dim

        self.G = np.identity(self.dim)  # .dim is the dimension
        self.U = np.zeros(self.dim)  # vector of averages
        self.A = np.zeros(self.dim)  # vectors of actions taken so far weighted by the reward
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)

        # give each arm a vector mean
        for arm in arm_object_list:
            arm.mean_vec = vector_mean

    def pullArm(self, arm):
        self.T[arm] += 1
        self.timestep += 1  # update timesetp

        # self.G is the sum of the products X X^T of the arm pulled in each round X
        # self.arm_vecs[arm] is the arm vector for a given [arm]
        self.G += np.outer(self.arm_vecs[arm], self.arm_vecs[arm])

        # self.A is the sum of the arm vector pulled so far, scaled by the reward each pull resulted in
        self.A += self.arm_vecs[arm] * self.arms[arm].pull()

        # self.U is the "most likely" value of the mean vector; it is calculated based on least squares
        # update the reward approximation based on the inverse of G times the scaled sum of chosen arms
        self.U = np.dot(inv(self.G), self.A)

        return None

    def giveRegret(self):
        return np.amax(np.dot(self.arm_vecs, self.mean)) * self.timestep[0] - np.dot(
            self.T,
            np.dot(self.arm_vecs, self.mean)
        )


    def reset(self):
        self.T = np.zeros(self.n_arms)
        self.G = np.identity(self.dim)  # .dim is the dimension
        self.U = np.zeros(self.dim)  # vector
        self.A = np.zeros(self.dim)  # vectors of actions taken so far weighted by the reward
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        return None


    def fullInfo(self):
        print("\n" + "+" + "-"*105 + "+")
        print("| Mean Vector: {}\n|".format(self.mean))
        print("| Gram Matrix: \n{}\n| ".format(self.G))
        for arm in range(self.n_arms):
            print("| Arm {0}: {1}, confidence ({2:f}), called ({3}) times\n| ".format(
                arm,
                self.arms[arm].arm_vec,
                self.U_conf[arm],
                self.T[arm])
            )
        print("Mean Reward Vector: {}".format(self.U))
        print("|\n| {0} arms,  timestep ({1} / {2})".format(
            self.n_arms,
            self.timestep[0],
            self.horizon[0])
        )
        print("+" + "-"*105 + "+" + "\n")
        return None



class AdvBandit:
    def __init__(self, arm_object_list):
        self.arms = arm_object_list
        self.seq = np.array([arm.reward_vec for arm in self.arms])
        self.n_arms, self.horizon = self.seq.shape
        self.horizon = [self.horizon] * self.n_arms

        self.U = np.ones(self.n_arms)
        self.W = np.ones(self.n_arms)  # weights, updated by algorithm
        self.T = np.zeros(self.n_arms)
        self.pulled = np.full(self.seq.shape, False, dtype=bool)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)


    def pullArm(self, arm):
        self.T[arm] += 1
        self.timestep += 1
        self.pulled[arm][self.timestep-1] = True

    def giveRegret(self):
        values = self.pulled * self.seq
        total_reward = np.sum(values)
        # the slice shortens the row by truncating any columns beyond the timestep
        best_arm_reward = np.amin(np.sum(self.seq[:,:self.timestep[0]], axis=1))  # compute the best arm loss
        return total_reward - best_arm_reward


    def reset(self):
        self.W = np.ones(self.n_arms)  # weights, updated by algorithm
        self.T = np.zeros(self.n_arms)
        self.pulled = np.full(self.seq.shape, False, dtype=bool)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        self.loss_approx = np.zeros(self.n_arms, dtype=np.int)

    def fullInfo(self):
        print("\n" + "+" + "-"*85 + "+")

        for arm in range(self.n_arms):
            print("| Arm {0}: weight ({1:f}), called ({2}) times".format(
                arm,
                self.W[arm],
                self.T[arm]))
        print("|\n| {0} arms, regret ({1}), timestep ({2} / {3})"
            .format(
                self.n_arms,
                self.giveRegret(),
                self.timestep[0],
                self.horizon[0]))

        print("+" + "-"*85 + "+" + "\n")
        return None


class Bandit:
    def __init__(self, bandit_arm):
        self.seq, self.distr = bandit_arm.give()
        self.info = bandit_arm.info  # the pre-calculated reward sequences
        self.pulled = np.full_like(self.seq, False, dtype=bool)  # tells which arms were pulled in each round (boolean array)
        self.timestep = 0
        self.n_arms, self.horizon = self.seq.shape
        self.T = np.zeros(self.n_arms, dtype=int)
        return None

    def giveRegret(self):
        values = self.pulled * self.distr
        total_reward = np.sum(values)
        # the slice shortens the row by truncating any columns beyond the timestep
        best_arm_reward = np.amin(np.sum(self.distr[:,:self.timestep], axis=1))  # compute the best arm loss
        return total_reward - best_arm_reward

    def pullArm(self, arm):
        self.T[arm] += 1
        self.pulled[arm][self.timestep] = True
        self.timestep += 1
        return None

    def fullInfo(self):
        print("\n" + "+" + "-"*85 + "+")
        for i, sub_dict in enumerate(self.info):
            print('Arm {}, pulled {} times: {}'.format(i, self.T[i], sub_dict))
        print('Regret {}, Timestep {} / {}'.format(self.giveRegret(), self.timestep, self.horizon))
        print(self.seq)
        print(self.pulled)
        print(self.distr)
        print("+" + "-"*85 + "+" + "\n")

class Algorithm:
    """
    The Algorithm class specifies the behaviour of an algorithm.

    Positional Arguments:
        * list of arm objects
        * vector mean
    Keyword Arguments:
        * algtype: the algorithm function you want to use
        * kwargs: other necessary objects depending on the algorithm
    Attributes:
        * var_dict: stores kwargs passed
    Methods:
        * giveArm(arm): provides the 'best' arm based on the algorithm
    Caveats:
        * you must pass it a bandit object
    """
    def __init__(self, **variables):
        self.var_dict = variables

    def giveArm(self):
        return self.var_dict['algtype'](self.bandit, self.var_dict)
